-----------------------------------------------------------------------------------------------
-- Client Lua Script for NameView
-- Copyright Celess. Portions (C) NCsoft via Nameplates. All rights reserved
-----------------------------------------------------------------------------------------------

require "Window"
require "ChallengesLib"
require "Unit"
require "GameLib"
require "Apollo"
require "PathMission"
require "Quest"
require "Episode"
require "math"
require "string"
require "DialogSys"
require "PublicEvent"
require "PublicEventObjective"
require "CommunicatorLib"
require "GroupLib"
require "PlayerPathLib"
require "GuildLib"
require "GuildTypeLib"
require "AccountItemLib"
require "MatchingGameLib"

-----------------------------------------------------------------------------------------------
-- NameView Module Definition
-----------------------------------------------------------------------------------------------
local _Version,_VersionMinor = 70923,0
local _AddonName,_VersionWindow = "NameView", 3
local NameView = {}

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------

local _eSaveLevel = GameLib.CodeEnumAddonSaveLevel

local _defaultSettings = {		--type: 0 custom, 1 single check, 2 slider, 3 color, 4 radio, 5 text, 6 number, 7 check parts, 8 signal, 9 drop
	-- Normal
	-- display
		-- CheckVisibilityOptions
	["bShowObjectives"] = { default=true, nControlType=1, strControlName="MainShowObjectives" },
	["bShowVendor"] = { default=true, nControlType=1, strControlName="MainShowVendors" },
	["bShowTaxi"] = { default=true, nControlType=1, strControlName="MainShowTaxis" },
	["bShowNeutral"] = { default=false, nControlType=1, strControlName="MainShowDisposition_2" },
	["bShowHostile"] = { default=true, nControlType=1, strControlName="MainShowDisposition_1" },
	["bShowMyPlate"] = { default=false, nControlType=1, strControlName="MainShowMine" },
	["bShowOrganization"] = { default=true, nControlType=1, strControlName="MainShowOrganization" },
	["bShowGroup"] = { default=true, nControlType=1, strControlName="MainShowGroup" },
	["bShowFriendly"] = { default=false, nControlType=1, strControlName="MainShowDisposition_3", fnCallback="OnEnablerButtonChecked" },
	["bShowFriendlyPlayer"] = { default=true, nControlType=1, strControlName="MainShowDisposition_FriendlyPlayer" },
		-- CheckDrawDistance
	["nMaxRange"] = { default=70.0, nControlType=0 },
		-- CheckVisibilityOptions
	["bUseOcclusion"] = { default=true, nControlType=1, strControlName="MainUseOcclusion", fnCallback="OnOcclusionCheck" },	-- needs to have a real reset value
	-- individual nameplates
		-- DrawName
	["bShowName"] = { default=true, nControlType=1, strControlName="IndividualShowName", fnCallback="OnSettingNameChanged" },
		-- DrawCastBar
	["bShowCastBar"] = { default=false, nControlType=1, strControlName="IndividualShowCastBar" },
		-- UpdateVisibility
	["bShowCertainDeath"] = { default=true, nControlType=1, strControlName="IndividualShowCertainDeath" },
		-- DrawName, DrawGuild
	["bShowAffiliation"] = { default=true, nControlType=1, strControlName="IndividualShowAffiliation", fnCallback="OnSettingTitleChanged" },
		-- DrawRewards
	["bShowRewards"] = { default=true, nControlType=1, strControlName="IndividualShowRewardIcons", fnCallback="OnRequestRewardsUpdate" },
		-- CheckVisibilityOptions
	["bShowDuringSpeech"] = { default=true, nControlType=1, strControlName="IndividualShowDuringSpeech" },
	-- health bar
		-- DrawVulnerable, DrawHealthBar
	["bShowHealth"] = { default=false, nControlType=4, strControlName="MainShowHealthBarAlways", strControlGroup="MainShowHealthBarGroup", fnCallback="OnSettingHealthChanged" },
	["bShowHealthDamaged"] = { default=true, nControlType=4, strControlName="MainShowHealthBarDamaged", strControlGroup="MainShowHealthBarGroup", fnCallback="OnSettingHealthChanged" },
	["bShowHealthNever"] = { default=false, nControlType=4, strControlName="MainShowHealthBarNever", strControlGroup="MainShowHealthBarGroup", fnCallback="OnSettingHealthChanged" },
	-- reward icons
		-- UpdateRewards
	["bShowIconQuests"] = { default=true, nControlType=1, strControlName="ShowRewardTypeQuest", fnCallback="OnRequestRewardsUpdate" },
	["bShowIconMissions"] = { default=true, nControlType=1, strControlName="ShowRewardTypeMission", fnCallback="OnRequestRewardsUpdate" },
	["bShowIconFriends"] = { default=true, nControlType=1, strControlName="ShowRewardTypeFriend", fnCallback="OnRequestRewardsUpdate" },
	["bShowIconPublicEvents"] = { default=true, nControlType=1, strControlName="ShowRewardTypePublicEvent", fnCallback="OnRequestRewardsUpdate" },
	["bShowIconChallenges"] = { default=true, nControlType=1, strControlName="ShowRewardTypeChallenge", fnCallback="OnRequestRewardsUpdate" },
	["bShowIconRivals"] = { default=true, nControlType=1, strControlName="ShowRewardTypeRival", fnCallback="OnRequestRewardsUpdate" },

	-- Targeted
	-- details
		-- DrawTargeting
	["bShowTargetSelection"] = { default=true, nControlType=1, strControlName="TargetShowSelection" },
		-- DrawName
	["bShowTargetName"] = { default=true, nControlType=1, strControlName="TargetShowName", fnCallback="OnSettingNameChanged" },
		-- DrawGuild
	["bShowTargetAffiliation"] = { default=true, nControlType=1, strControlName="TargetShowAffiliation", fnCallback="OnSettingTitleChanged" },
		-- DrawVulnerable, DrawHealthBar
	["bShowTargetHealth"] = { default=true, nControlType=1, strControlName="TargetShowHealth", fnCallback="OnSettingHealthChanged" },
		-- DrawRewards
	["bShowTargetRewards"] = { default=true, nControlType=1, strControlName="TargetShowRewards"},
		-- DrawCastBar
	["bShowTargetCastBar"] = { default=true, nControlType=1, strControlName="TargetShowCastBar" },
	-- non-targeted nameplates in combat
		-- CheckVisibilityOptions
	["bShowInCombat"] = { default=true, nControlType=1, strControlName="MainHideInCombatOff" },

	-- Advanced
	-- display
	["bShowEnemysUnits"] = { default=true, nControlType=1, strControlName="AdvShowEnemysUnits" },
	-- style
	["bFontLargeName"] = { default=false, nControlType=1, strControlName="AdvFontNameLarge", fnCallback="OnSettingFontChanged" },
	["bFontLargerName"] = { default=true, nControlType=1, strControlName="AdvFontNameLargerer", fnCallback="OnSettingFontChanged" },
	["bColorLightGuild"] = { default=true, nControlType=1, strControlName="AdvColorGuildLight", fnCallback="OnSettingFontChanged" },
	["bColorLightEnemy"] = { default=false, nControlType=1, strControlName="AdvColorEnemyLight", fnCallback="OnSettingFontChanged" },
	["bPerspectiveScale"] = { default=false, nControlType=1, strControlName="AdvPerspectiveScale", fnCallback="OnEnablerButtonChecked" },
	["bPerspectiveOpacity"] = { default=false, nControlType=1, strControlName="AdvPerspectiveOpacity", fnCallback="OnEnablerButtonChecked" },
	["bPerspectiveCombat"] = { default=true, nControlType=1, strControlName="AdvPerspectiveCombat" },
	["bPerspectiveCombatants"] = { default=false, nControlType=1, strControlName="AdvPerspectiveCombatants" },
	-- individual nameplates
	["bShowText"] = { default=true, nControlType=1, strControlName="AdvShowText", fnCallback="OnEnablerButtonChecked" },
	["bShowMaxHealthText"] = { default=true, nControlType=1, strControlName="AdvShowMaxHealthText" },
	["bShowShieldText"] = { default=true, nControlType=1, strControlName="AdvShowShieldText" },
	["bShowMarker"] = { default=true, nControlType=1, strControlName="AdvShowMarker" },
	["bShowVulnBar"] = { default=true, nControlType=1, strControlName="AdvShowVulnBar" },
	["bShowFirst"] = { default=true, nControlType=1, strControlName="AdvShowFirst", fnCallback="OnSettingNameChanged" },
	["bShowLast"] = { default=true, nControlType=1, strControlName="AdvShowLast", fnCallback="OnSettingNameChanged" },
	["bShowFullNpc"] = { default=true, nControlType=1, strControlName="AdvShowFullNpc", fnCallback="OnSettingNameChanged" },
	["bShowTitle"] = { default=true, nControlType=1, strControlName="AdvShowTitle", fnCallback="OnSettingTitleChanged" },
	["bShowGuild"] = { default=true, nControlType=1, strControlName="AdvShowGuild", fnCallback="OnSettingTitleChanged" },
	-- targeted
	["bShowTargetText"] = { default=true, nControlType=1, strControlName="TargetShowText", fnCallback="OnEnablerButtonChecked" },
	["bShowTargetMaxHealthText"] = { default=true, nControlType=1, strControlName="TargetShowMaxHealthText" },
	["bShowTargetShieldText"] = { default=true, nControlType=1, strControlName="TargetShowShieldText" },
	["bShowTargetMarker"] = { default=true, nControlType=1, strControlName="TargetShowMarker" },
	["bShowTargetVulnBar"] = { default=true, nControlType=1, strControlName="TargetShowVulnBar" },
	["bTargetPerspective"] = { default=false, nControlType=1, strControlName="TargetPerspective" },
	["bShowTargetFull"] = { default=true, nControlType=1, strControlName="TargetShowFull", fnCallback="OnSettingNameChanged" },

	-- Other

	sSettingsName = "Default",
	nSettingsVersion = 1,
	tSettingsCustom = {	},
}

local _resetSettings = {
	{
		sSettingsName = "Default",
		sSettingsTooltip = "Default settings. These are the same as the current Carbine settings for Nameplates.",
	}, {
		sSettingsName = "Updated",
		sSettingsTooltip = "A more progressive style than Default, with a few of the newer NameView features and refinements. Changes: Range changed from 70 to 90 meters, Cast Bars enabled, Larger Name Font enabled, Max-Health Text disabled, Title disabled, Perspective Opacity enabled.",
		nMaxRange = 90,						-- Normal
		bShowCastBar = true,
		bFontLargeName = true,				-- Advanced
		bFontLargerName = false,
		bColorLightEnemy = true,
		bPerspectiveOpacity = true,
		bShowMaxHealthText = false,
		bShowTitle = false,
	}, {
		sSettingsName = "Perspective",
		nMaxRange = 125,					-- Normal
		bShowCastBar = true,
		bFontLargeName = true,				-- Advanced
		bColorLightEnemy = true,
		bPerspectiveScale = true,
		bPerspectiveOpacity = true,
		bShowMaxHealthText = false,
		bShowShieldText = false,
		bShowTitle = false,
		bShowTargetMarker = false,
	}, {
		sSettingsName = "Massive",
		bShowNeutral = true,				-- Normal
		bShowMyPlate = true,
		bShowFriendly = true,
		nMaxRange = 150,
		bUseOcclusion = false,
		bShowCastBar = true,
		bShowHealth = true,
		bShowHealthDamaged = false,
		bFontLargeName = true,				-- Advanced
		bColorLightEnemy = true,
		bPerspectiveScale = true,
		bPerspectiveOpacity = true,
		bPerspectiveCombatants = true,
	}, {
		sSettingsName = "Custom",
		sSettingsTooltip = "Custom defaults. Use the Save button at any time to save current settings to this Custom preset. Load them again by choosing this Custom preset from the list.",
	}, {
		sSettingsName = "Durina",
		bShowMyPlate = true,				-- Normal
		nMaxRange = 100,
		bShowCastBar = true,
		bShowAffiliation = false,
		bShowEnemysUnits = false,			-- Advanced
		bFontLargeName = true,
		bFontLargerName = false,
		bColorLightEnemy = true,
		bPerspectiveScale = true,
		bPerspectiveOpacity = true,
		bShowMarker = false,
		bShowMaxHealthText = false,
		bShowShieldText = false,
		bShowLast = false,
		bShowTargetMarker = false,
	}, {
		sSettingsName = "Pancakes",
		bShowVendor = false,				-- Normal
		bShowNeutral = true,
		bShowOrganization = false,
		bShowGroup = false,
		nMaxRange = 75,
		bShowCastBar = true,
		bShowAffiliation = false,
		bShowHealth = true,
		bShowHealthDamaged = false,
		bShowIconFriends = false,
		bShowIconRivals = false,
		bShowTargetAffiliation = false,		-- Targeted
		bShowEnemysUnits = false,			-- Advanced
		bFontLargeName = true,
		bColorLightEnemy = true,
		bPerspectiveScale = true,
		bPerspectiveCombatants = true,
		bShowMaxHealthText = false,
		bShowShieldText = false,
		bShowFullNpc = false,
		bShowTitle = false,
		bShowGuild = false,
		bShowTargetMaxHealthText = false,
		bShowTargetShieldText = false,
	},
}
for _,v in pairs(_resetSettings) do
	_resetSettings[v.sSettingsName:lower()] = v
end

local _eDisposition = Unit.CodeEnumDisposition
local _eStateVulnerability = Unit.CodeEnumCCState.Vulnerability
local _eRewardInfoType = Unit.CodeEnumRewardInfoType

local _eTintType = {
	Hostile = _eDisposition.Hostile,
	Neutral = _eDisposition.Neutral,
	Friendly = _eDisposition.Friendly,
	Unknown = _eDisposition.Unknown,
	Vulnerable = _eDisposition.Unknown + 1,
}

local _tTextColors = {
	[_eDisposition.Hostile]  = ApolloColor.new("DispositionHostile"),
	[_eDisposition.Neutral]  = ApolloColor.new("DispositionNeutral"),
	[_eDisposition.Friendly] = ApolloColor.new("DispositionFriendly"),
	[_eDisposition.Unknown]  = ApolloColor.new("DispositionNeutral"),
}

local _tHealthBar = {
	[_eTintType.Hostile] = "CRB_Nameplates:sprNP_RedProg",
	[_eTintType.Neutral] = "CRB_Nameplates:sprNP_YellowProg",
	[_eTintType.Friendly] = "CRB_Nameplates:sprNP_GreenProg",
	[_eTintType.Unknown] = "CRB_Nameplates:sprNP_YellowProg",
	[_eTintType.Vulnerable] = "CRB_Nameplates:sprNP_PurpleProg",
}

local _tTargetBar = {
	[_eTintType.Hostile] = "CRB_Nameplates:sprNP_BaseSelectedRed",
	[_eTintType.Neutral] = "CRB_Nameplates:sprNP_BaseSelectedYellow",
	[_eTintType.Friendly] = "CRB_Nameplates:sprNP_BaseSelectedGreen",
	[_eTintType.Unknown] = "CRB_Nameplates:sprNP_BaseSelectedYellow",
	[_eTintType.Vulnerable] = "CRB_Nameplates:sprNP_BaseSelectedPurple",
}

--local _tTargetPrimary = {				--FIXME: i think this was inadvertantly removed from use
--	[_eDisposition.Hostile] = "sprNp_Target_HostilePrimary",
--	[_eDisposition.Neutral] = "sprNp_Target_NeutralPrimary",
--	[_eDisposition.Friendly] = "sprNp_Target_FriendlyPrimary",
--	[_eDisposition.Unknown] = "sprNp_Target_NeutralPrimary",
--}

local _tTargetSecondary = {
	[_eDisposition.Hostile] = "sprNp_Target_HostileSecondary",
	[_eDisposition.Neutral] = "sprNp_Target_NeutralSecondary",
	[_eDisposition.Friendly] = "sprNp_Target_FriendlySecondary",
	[_eDisposition.Unknown] = "sprNp_Target_NeutralSecondary",
}

local _nHealthRed = 0.3
local _nHealthYellow = 0.5

--TODO: break these out onto options
local _crGroup				= ApolloColor.new("DispositionFriendlyUnflaggedDull")
local _crGuild_Dark			= ApolloColor.new("DispositionGuildmateUnflagged")
local _crGuild_Light		= ApolloColor.new("FFB970DB")	--FFAF5CD6 10% -- lightened about 20%. the regular purple color is very hard to read as its very dark against the black outline
local _crGuild				= _crGuild_Dark					--NOTE: non-static var.  the light color is also much closer to a matching color vs the other game text colors
--local _crWarparty			= ApolloColor.new("DispositionGuildmateUnflagged")
local _crAllyUnflag			= ApolloColor.new("DispositionFriendlyUnflagged")
local _crAllyUnflag_Light	= _crTextChange	--ApolloColor.new("DispositionFriendlyUnflagged")
local _crAlly				= _tTextColors[_eDisposition.Friendly]
local _crEnemyUnflag 		= _tTextColors[_eDisposition.Neutral]
local _crEnemy_Dark 		= _tTextColors[_eDisposition.Hostile]
local _crEnemy_Light 		= ApolloColor.new("FFFFF322A") --FA3A26") --f7211d") --FA2521")		--FFFF211C") --FFFF3633" 20% --FFFF1D1A 10% ffff0500-- lightened about 20%. the regular red color is very hard to read as its very dark against the black outline
local _crEnemy_LightLight 	= ApolloColor.new("FFF75544") --FA3A26") --f7211d")	--FA2521")		--FFFF211C") --FFFF3633" 20% --FFFF1D1A 10% ffff0500-- lightened about 20%. the regular red color is very hard to read as its very dark against the black outline
local _crEnemy				= _crEnemy_Dark
local _crEnemyMismatch		= ApolloColor.new("DispositionPvPFlagMismatch")
local _crDead 				= ApolloColor.new("DeathGrey")
local _crDefaultTagged		= ApolloColor.new("DeathGrey")

local _sFont_Default	= "Nameplates"			-- CRB_Header10_O is "Nameplates"
local _sFont_Large		= "CRB_Interface10_BO"
local _sFont_Larger		= "CRB_Interface11_BO"
local _sFont_Large85	= "CRB_Interface10_BO"
local _sFont_Larger85	= "CRB_Interface11_BO"
local _sFont_Large75	= "CRB_Interface11_BO"
local _sFont_Larger75	= "CRB_Interface12_BO"
local _sFont_Large70	= "CRB_Header12_O"		-- panic values. very very common, people often use 1280 and .75 ui scale to play the game at all
local _sFont_Larger70	= "CRB_Header13_O"
local _sFont			= _sFont_Default
local _sGuildFont		= "CRB_Interface9_BO"

--local _csScalingHex	= "ffffbf80"
local _crScaling 		= CColor.new(1.0, 191/255, 128/255, 0.7)		--  "ffffbf80"

local _nConOffset = 5
local _nConMin = 1
local _nConMax = 9

local _arConColor = {						-- level differential color
	[1] = ApolloColor.new("ConTrivial"),	-- -4
	[2] = ApolloColor.new("ConInferior"),	-- -3
	[3] = ApolloColor.new("ConMinor"),
	[4] = ApolloColor.new("ConEasy"),
	[5] = ApolloColor.new("ConAverage"),
	[6] = ApolloColor.new("ConModerate"),
	[7] = ApolloColor.new("ConTough"),
	[8] = ApolloColor.new("ConHard"),
	[9] = ApolloColor.new("ConImpossible"),
}

local _sHighLevelSprite = "CRB_Nameplates:sprNP_HighLevel"
local _sPvpTargetSprite = "IconSprites:Icon_Windows_UI_CRB_Marker_Crosshair"

local _arPathSprites = {
	[PlayerPathLib.PlayerPathType_Soldier] 		= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSol",
	[PlayerPathLib.PlayerPathType_Settler] 		= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSet",
	[PlayerPathLib.PlayerPathType_Scientist] 	= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSci",
	[PlayerPathLib.PlayerPathType_Explorer] 	= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathExp",
}

local _floor = math.floor
local _sqrt = math.sqrt
local _abs = math.abs
local _format = string.format
local _len = string.len
local _tsort = table.sort
local _byte = string.byte
local _char = string.char

local _nTargetMax 			= 200			-- max distance to show if targeted. same as client max range for showing targeted models
local _nRangeMax 			= 150			-- max distance to show normally. same as client max range for showing most non-targeted models
--local _nTargetMaxSq 		= 40000			-- target max squared
local _nRangeMaxSq 			= 22500			-- range max squared. used as a psudo-value to provide vanishing point range for Perspective options
local _nNearClip_Default	= 0.007			-- default, ~10m until first step at 1.0 curve and 1.0 trim and ~1366 or better resolutions
local _nNearClip_85			= 0.021			-- larger dead zone for lower ui scale and resolution H and W values
local _nNearClip_75			= 0.035			-- ~30m at 1.0 and 1.0 and .75 and 1280x. much larger dead zone for very low ui scale and resolution H and W values
local _nNearClip			= _nNearClip_	-- near clipping for special effects. ratio of unit from player to vanish point
local _nFarClip				= 0.260			-- far clipping for special effects. ratio of unit from player to vanish point
local _nVanishCurve 		= 0.96			-- perspecitve scaling curve. less is shallower curve, but much higher values that need trim
local _nVanishTrim 			= 1.06			-- perspecitve scaling trim
local _nOpacityCurve		= 1.10			-- perspecitve opacity curve. rebalance opacity with scale curve
local _nOpacityTrim			= 0.95			-- perspecitve opacity trim

local _nPoolLimit			= 100			-- the pool max size
local _nPoolMinLimit		= 15			-- the minimum window pool size
local _nTimerSpeed			= 0.25
local _nAniSpeed			= 2					-- the number of animation updates
local _nReclamationAge		= 25 * _nAniSpeed 	-- the number visi test ticks before window is reclaimed

local _LTargetFrame_HealthText = Apollo.GetString("TargetFrame_HealthText")
local _LTargetFrame_HealthShieldText = Apollo.GetString("TargetFrame_HealthShieldText")
local _LNameplates_GuildDisplay = Apollo.GetString("Nameplates_GuildDisplay")

local _tPixie = {
	strSprite = "", cr = "",
	strText = nil, flagsText = {},
	loc = { fPoints = { 0, 0, 0, 0 }, nOffsets = { 0, 0, 0, 0 } }
}

local function _dummy() end
local print = _dummy

local function _print(...)
	local s = ""
	for i = 1, select("#",...) do
		s = s .. tostring(select(i,...)) .. " "
	end
	Print(s)
end

local function DeepCopy(o)
	if type(o) ~= 'table' then return o end

	local copy = {}
	for k,v in next, o do copy[k] = DeepCopy(v) end
	return copy
end

local function SettingsCopy(o)
	local copy = {}
	for k,v in next, o do
		if type(v) == "table" and v.nControlType then v = v.default end
		copy[k] = DeepCopy(v)
	end
	return copy
end

local function PostChat(s)
	ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, s)
end

local _tShortFormatWhole = {
	[1] = Apollo.GetString("TargetFrame_ShortNumberWhole"),		-- thousands
	[2] = Apollo.GetString("TargetFrame_MillionsNumberWhole"),
	[3] = Apollo.GetString("TargetFrame_BillionsNumberWhole")
}

local _tShortFormatFloat = {
	[1] = Apollo.GetString("TargetFrame_ShortNumberFloat"),		-- thousands
	[2] = Apollo.GetString("TargetFrame_MillionsNumberFloat"),
	[3] = Apollo.GetString("TargetFrame_BillionsNumberFloat")
}

local function ShortNumber( n )							-- replaces FormatNumber/BigNumber, short number packed with thousands
	local i = 0											--   place indicator, for fast comparisons between frames.
	while n >= 1000 and i < 3 do						--   format: 54321 as int, where 1 is thousands root, and 543.2 is short form
		n = n / 1000									-- use with matching FormatNumber
		i = i + 1
	end
	return _floor((n - n % .1) * 100 + i)				-- floor to clip float errors
end

local function ShortRound( n )							-- replaces FormatNumber/BigNumber, short number packed with thousands rounded
	local i = 0											--   place indicator, for fast comparisons between frames.
	while n >= 1000 and i < 3 do						--   format: 54321 as int, where 1 is thousands root, and 543.2 is short form
		n = n / 1000									-- use with matching FormatNumber
		i = i + 1
	end
	n = n + 0.05; return _floor((n - n % .1) * 100 + i)	-- floor to clip float errors
end

local function FormatNumber( n )
	local i = n % 10									-- thousands root
	n = (n - i) / 100									-- short number form
	return i == 0 and tostring(n)
		or String_GetWeaselString((n % 1 == 0 and _tShortFormatWhole or _tShortFormatFloat)[i], n)
end

local function ShortTime( n )							-- replaces ConvertSecondsToTimer,
	local s = _floor(n)									--   for fast comparisons between frames
	return s											-- use with matching FormatTime
end

local function FormatTime( n )
	local s = n % 60
	local m = _floor(n % 3600 / 60)
	local h = _floor(n / 3600)
	return (h == 0  and "" or h .. ":") .. m .. ":" .. (s < 10 and (0 .. s) or s)
end


-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------

function NameView:new(o)
	o = self
	--o = o or {}
	--setmetatable(o, self)
	--self.__index = self

	o.L = o.L or _L or {}
	setmetatable(o.L, { __index = function(t,k) return tostring(k) end })

	o._AddonName = _AddonName
	o._defaultSettings = _defaultSettings
	o._resetSettings = _resetSettings
	o._print = _print

	o.arShowPlates = {}
	o.arNameplates = {}
	o.arUnit2Plate = {}

	o.arPreloadUnits = {}
	o.arPlatePool = {}
	o.arWindowPool = {}

	--o.unitPlayer = nil
	--o.bPlayerInCombat = nil
	--o.tPlayerGuild = nil
	--o.tPlayerWarparty = nil
	--o.bRequestRewards = nil

	self.nFrame = 0
	self.nAge = 0
	o.arAniPlates = {}

	return o
end

function NameView:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = Apollo.GetString("CRB_Nameplates")		-- "NameView"
	local tDependencies = { "RewardIcons" }		-- { "Tooltips", "RewardIcons" }
	Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

function NameView:NewLocale(lc, default)
	local L,s = self.L, Apollo.GetString(1)
	if (s == "Annuler" and "frFR" or (s == "Abbrechen" and "deDE" or "enUS")) == lc then return L,self end

	if default then
		self.Ldefault = setmetatable({}, getmetatable(L))
		setmetatable(L, { __index = self.Ldefault })
		return self.Ldefault,self
	end
end

function NameView:OnDependencyError(strDependency, strError)
	return true
end


-----------------------------------------------------------------------------------------------
-- NameView OnLoad
-----------------------------------------------------------------------------------------------

function NameView:OnLoad()
	self.onXmlDocLoadedCalled = false

	self.userSettings = SettingsCopy(_defaultSettings)

	-- create form
	self.xmlDoc = XmlDoc.CreateFromFile("NameView.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)

	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)	-- unit events can get created before docload

	Apollo.RegisterEventHandler("WindowManagementReady", "OnWindowManagementReady", self)
	Apollo.RegisterEventHandler("NameView_ShowMenu", "OnShowMenu", self)
	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
	Apollo.RegisterSlashCommand("Nameplates", "OnCommand", self)
	Apollo.RegisterSlashCommand("nameplates", "OnCommand", self)
	Apollo.RegisterSlashCommand("NameView", "OnCommand", self)
	Apollo.RegisterSlashCommand("nameview", "OnCommand", self)
end

function NameView:OnWindowManagementReady()
	Event_FireGenericEvent("WindowManagementRegister", {strName=Apollo.GetString("Nameplates_Options"), nSaveVersion=_VersionWindow})
end

function NameView:OnInterfaceMenuListHasLoaded()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", _AddonName, --Apollo.GetString("CRB_Nameplates"),
		{ "NameView_ShowMenu", "", "CRB_ActionBarFrameSprites:sprResourceBar_Sprint_RunIconBlue" })
end


-----------------------------------------------------------------------------------------------
-- NameView OnDocLoaded
-----------------------------------------------------------------------------------------------

function NameView:OnDocLoaded()
	if self.xmlDoc == nil or not self.xmlDoc:IsLoaded() then
		return
	end

	Event_FireGenericEvent("OneVersion_ReportAddonInfo", _AddonName, 1, _Version, _VersionMinor)

	-- Load the window

	local w = Apollo.LoadForm(self.xmlDoc, "Nameplate", nil, self)					-- cache defaults
	self.wndPlateDummy = w
	self.nPlateLeft, self.nPlateTop, self.nPlateRight, self.nPlateBottom = w:GetAnchorOffsets()
	self.nGuildHeight = w:FindChild("Guild"):GetHeight()
	self.nHealthLeft, self.nHealthTop, self.nHealthRight, self.nHealthBottom = w:FindChild("Bars"):GetAnchorOffsets()
	self.rewardsDummy = w:FindChild("Rewards")

	self:SetPlayerGuilds()

	-- Apollo.RegisterEventHandler("UnitCreated", 					"OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed",				"OnUnitDestroyed", self)
	Apollo.RegisterEventHandler("NextFrame",					"OnFrame", self)

	Apollo.RegisterEventHandler("UnitTextBubbleCreate",			"OnUnitTextBubbleToggled", self)
	Apollo.RegisterEventHandler("UnitTextBubblesDestroyed",		"OnUnitTextBubbleToggled", self)
	Apollo.RegisterEventHandler("TargetUnitChanged",			"OnTargetUnitChanged", self)
	Apollo.RegisterEventHandler("UnitEnteredCombat",			"OnEnteredCombat", self)
	Apollo.RegisterEventHandler("UnitNameChanged",				"OnUnitNameChanged", self)
	Apollo.RegisterEventHandler("UnitTitleChanged",				"OnUnitTitleChanged", self)
	Apollo.RegisterEventHandler("PlayerTitleChange",			"OnPlayerTitleChanged", self)
	Apollo.RegisterEventHandler("UnitGuildNameplateChanged",	"OnUnitGuildNameplateChanged",self)
	Apollo.RegisterEventHandler("UnitLevelChanged",				"OnUnitLevelChanged", self)
	Apollo.RegisterEventHandler("UnitMemberOfGuildChange",		"OnUnitMemberOfGuildChange", self)
	Apollo.RegisterEventHandler("GuildChange",					"OnGuildChange", self)
	Apollo.RegisterEventHandler("UnitGibbed",					"OnUnitGibbed", self)
	Apollo.RegisterEventHandler("ResolutionChanged",			"OnResolutionChanged", self)

	local t = {
		"QuestObjectiveUpdated", "QuestStateChanged", "ChallengeAbandon", "ChallengeLeftArea", "ChallengeFailTime",
		"ChallengeFailArea", "ChallengeActivate", "ChallengeCompleted", "ChallengeFailGeneric", "PublicEventObjectiveUpdate",
		"PublicEventUnitUpdate", "PlayerPathMissionUpdate", "FriendshipAdd", "FriendshipPostRemove", "FriendshipUpdate",
		"PlayerPathRefresh", "ContractObjectiveUpdated", "ContractStateChanged", "ChallengeUpdated",
	}
	for _,v in pairs(t) do
		Apollo.RegisterEventHandler(v, "OnRequestRewardsUpdate", self)
	end

	self.tUpdateTimer = ApolloTimer.Create(_nTimerSpeed, true, "OnUpdateTimer", self)

	self.onXmlDocLoadedCalled = true
	self:InitializeWindows()
end

function NameView:SetPlayerGuilds()
	self.tPlayerGuild = nil
	for _,v in next, GuildLib.GetGuilds() do
		if v:GetType() == GuildLib.GuildType_Guild then
			self.tPlayerGuild = v
			break
		end
	end
	--self.tPlayerGuild = nil				--NOTE: warparty color. disabled as carbine version is disabled
	--self.tPlayerWarparty = nil
	--for _,v in next, GuildLib.GetGuilds() do
	--	local eGuildType = v:GetType()
	--	if eGuildType == GuildLib.GuildType_Guild then
	--		self.tPlayerGuild = v
	--	elseif eGuildType == GuildLib.GuildType_WarParty then
	--		self.tPlayerWarparty = v
	--	end
	--end
end


-----------------------------------------------------------------------------------------------
-- NameView Functions
-----------------------------------------------------------------------------------------------

function NameView:OnFrame()
	--self.nFrame = self.nFrame + 1
	--local nTime = GameLib.GetTickCount()													--PROF:
	--self._frameCount = (self._frameCount or 0) + 1										--PROF:
	local o = self.userSettings

	for _,plate in next, self.arShowPlates do

		local unit = plate.unitOwner

		-- cast bar

		local bShow = plate.bShowCast and unit:ShouldShowCastBar()

		if bShow ~= plate.bShowCastBar and plate.castBar then		-- plate.X may be nil, thats fine
			plate.castBar:Show(bShow)
			plate.bShowCastBar = bShow
		end

		if bShow then
			self:DrawCast(plate)
		end

		-- vulnerable bar

		local v = plate.bShowVuln and unit:GetCCStateTimeRemaining(_eStateVulnerability) or 0
		bShow = v > 0

		if bShow ~= plate.bShowVulnBar and plate.vulnBar then
			plate.vulnBar:Show(bShow)
			plate.bShowVulnBar = bShow
		end

		if bShow then
			self:DrawVuln(plate, v)
		end

		-- health bar

		if plate.bShowHealthBar then
			self:DrawHealth(plate)
		end
	end
	--if self._frameCount % 30 == 1 then													--PROF:
	--	Print((self._frameCount or 0) .. ": " .. (GameLib.GetTickCount() - nTime))			--PROF:
	--end																					--PROF:
end

function NameView:DrawCast(plate)
	local unit = plate.unitOwner

	if not plate.castLabel then		--FIXME: carbine C side race condition
		return
	end

	local v = unit:GetCastName()
	if v ~= plate.sCastName then
		plate.castLabel:SetText(v)
		plate.sCastName = v
	end

	v = unit:GetCastDuration()
	if v ~= plate.nCastDuration then
		plate.castFill:SetMax(v)		-- set max before progress
		plate.nCastDuration = v
	end

	v = unit:GetCastElapsed()
	if v ~= plate.nCastElapsed then
		plate.castFill:SetProgress(v)
		plate.nCastElapsed = v
	end
end

function NameView:DrawVuln(plate, v)
	local unit = plate.unitOwner

	if not plate.vulnFill then		--FIXME: carbine C side race condition
		return
	end

	local u = unit:GetCCStateTotalTime(_eStateVulnerability)
	if u ~= plate.nVulnerableTimeMax then
		plate.vulnFill:SetMax(u)
		plate.nVulnerableTimeMax = u
	end

	plate.vulnFill:SetProgress(v)
end

function NameView:DrawHealth(plate)
	local unit = plate.unitOwner

	local nHealthMax = unit and unit:GetMaxHealth()

	-- color
	local bShow = plate.bShowVulnBar and _eTintType.Vulnerable or plate.eDisposition

	if bShow ~= plate.eHealthTintType then
		plate.health:SetFullSprite(_tHealthBar[bShow])					-- actual healbar bar
		plate.targetMarker:SetSprite(_tTargetBar[bShow])					-- health bar target arrows
		plate.targetArrow:SetSprite(_tTargetSecondary[plate.eDisposition])	-- top target arrow
		plate.eHealthTintType = bShow
	end

	-- simple
	bShow = not nHealthMax or plate.sUnitType == "Simple"
	if bShow ~= plate.bIsSimple then
		if bShow then
			plate.health:SetAnchorPoints(0, 0, 1, 1)
			plate.nHealth = nil						-- nHealth and nTotalMax contribute to health:SetAnchorOffsets
			plate.nTotalMax = nil
			plate.healthLabel:SetText("")
			plate.sHealthText = ""
			plate.nHealthMax = nil
		end
		plate.bIsSimple = bShow
	end
	if bShow then
--self._cSkip = (self._cSkip or 0) + 1													--PROF:
--Print("skip " .. tostring(nHealthMax) .. " " .. tostring(unit:GetName()))				--PROF:
		return
	end

	local nHealth     = (nHealthMax == 0 or unit:IsDead()) and 0 or unit:GetHealth()
	local nShieldMax  = unit:GetShieldCapacityMax() or 0
	local nShield     = nShieldMax == 0 and 0 or unit:GetShieldCapacity()
	local nAbsorbMax  = unit:GetAbsorptionMax() or 0
	local nAbsorb     = nAbsorbMax == 0 and 0 or unit:GetAbsorptionValue()	-- doesn't clear itself when absorb buff drops off
	local nHealAbsorb = unit:GetHealingAbsorptionValue() or 0
	local nHealMin    = unit:GetHealthFloor() or 0
	local nHealMax    = unit:GetHealthCeiling() or 0
	--local nHealAbsorb = (unit:GetHealingAbsorptionValue() or 0) + nHealthMax - nHealth
	--local nHealMin    = (unit:GetHealthFloor() or 0) + 1
	--local nHealMax    = (unit:GetHealthCeiling() or 0) + nHealthMax - 1
	local nTotalMax   = nHealthMax + nShieldMax + nAbsorbMax

	local bNewHealth = nHealth ~= plate.nHealth or nHealthMax ~= plate.nHealthMax
	local bNewShield = nShield ~= plate.nShield or nShieldMax ~= plate.nShieldMax
	local bNewAbsorb = nAbsorb ~= plate.nAbsorb or nAbsorbMax ~= plate.nAbsorbMax
	local bNewHeal = nHealAbsorb ~= plate.nHealAbsorb or nHealMin ~= plate.nHealMin or nHealMax ~= plate.nHealMax
	local bNewOption = plate.bNewTextOption

	if not bNewHealth and not bNewShield and not bNewAbsorb and not bNewHeal and not bNewOption then
		return
	end

	-- text
	if plate.bShowText and (bNewHealth or bNewShield or bNewOption) then

		local n,m,r = ShortRound(nHealth), ShortRound(nHealthMax), ShortRound(nShield)
		if n ~= plate.nHealthClip or m ~= plate.nHealthMaxClip or r ~= plate.nShieldClip or bNewOption then
			plate.nHealthClip, plate.nHealthMaxClip, plate.nShieldClip = n,m,r

			local s = FormatNumber(n)
			if plate.bShowMaxHealthText and nHealth ~= nHealthMax then
				s = String_GetWeaselString(_LTargetFrame_HealthText, s, FormatNumber(m))
			end
			if plate.bShowShieldText and nShieldMax > 0 and nShield > 0 then
				s = String_GetWeaselString(_LTargetFrame_HealthShieldText, s, FormatNumber(r))
			end

			if s ~= plate.sHealthText then
				plate.healthLabel:SetText(s)
				plate.sHealthText = s
			end
		end

		plate.bNewTextOption = nil
	end

	-- health
	local bShowHealth = nHealth > 0
	if nHealthMax ~= plate.nHealthMax or nTotalMax ~= plate.nTotalMax then
		local width = 1 / nTotalMax
		local h = nHealthMax * width
		local s = nShieldMax * width + h
		local a = nAbsorbMax * width + s

		h = _floor(h * 1000 + .01) / 1000; s = _floor(s * 1000 + .01) / 1000; a = _floor(a * 1000 + .01) / 1000	
		h = h < 1 and h or 1; s = s < 1 and s or 1; a = a < 1 and a or 1

		if h ~= plate.nHealthH or s ~= plate.nHealthS or a ~= plate.nHealthA then
			if s > h and (s - h) < .16 then					-- shield minimum 20%. nHealthRight is 103 total width. now 113 and more like 15%
				h = s - .16
			end
			if a > s and (a - s) < .08 then					-- absorb minimum, amp procs. else fill doesnt work correctly
				h = h - (.08 - (a - s)); s = a - .08;
			end
			plate.healAbsorb:SetAnchorPoints(0, 0, h, 1)
			plate.health:SetAnchorPoints(0, 0, h, 1)
			--plate.healMin:SetAnchorPoints(0, 0, h, 1)	-- children of health, layout is automatic
			--plate.healMax:SetAnchorPoints(0, 0, h, 1)
			plate.shield:SetAnchorPoints(h, 0, s, 1)
			plate.absorb:SetAnchorPoints(s, 0, a, 1)
			plate.nHealthH, plate.nHealthS, plate.nHealthA = h, s, a
		end

		plate.nHealthMax = nHealthMax
		plate.nTotalMax = nTotalMax
	end

	local nHealthMaxAbsorb = nHealthMax + nHealAbsorb

	-- heal absorb
	bShow = bShowHealth and nHealAbsorb > 0
	if bShow ~= plate.bShowHealAbsorb then
		plate.healAbsorb:Show(bShow)
		plate.bShowHealAbsorb = bShow
	end
	if bShow then
		local nHealAbsorbMark = nHealth + nHealAbsorb
		if nHealAbsorbMark ~= plate.nHealAbsorbMark or nHealthMaxAbsorb ~= plate.nHealthMaxAbsorb then
			plate.healAbsorb:SetMax(nHealthMaxAbsorb)						-- set max before progress or control breaks
			plate.healAbsorb:SetProgress(nHealAbsorbMark, nHealthMaxAbsorb * 4)
			plate.nHealAbsorbMark,plate.nHealthMaxAbsorb = nHealAbsorbMark, nHealthMaxAbsorb
		end
	end

	-- heal clamp
	bShow = bShowHealth and nHealMin > 0
	if bShow ~= plate.bShowHealMin then
		plate.healMin:Show(bShow)
		plate.bShowHealMin = bShow
	end
	if bShow and (nHealthMaxAbsorb ~= plate.nHealMinMax or nHealMin ~= plate.nHealMin) then
		plate.healMin:SetMax(nHealthMaxAbsorb)						-- set max before progress or control breaks
		plate.healMin:SetProgress(nHealMin)
		plate.nHealMinMax = nHealthMaxAbsorb
	end
	bShow = bShowHealth and nHealMax > 0 and nHealMax ~= nHealthMax 
	if bShow ~= plate.bShowHealMax then
		plate.healMax:Show(bShow)
		plate.bShowHealMax = bShow
	end
	if bShow and (nHealthMaxAbsorb ~= plate.nHealMaxMax or nHealMax ~= plate.nHealMax) then
		plate.healMax:SetMax(nHealthMaxAbsorb)						-- set max before progress or control breaks
		plate.healMax:SetProgress(nHealMax)
		plate.nHealMaxMax = nHealthMaxAbsorb
	end

	if bNewHeal then
		plate.nHealAbsorb, plate.nHealMin, plate.nHealMax = nHealAbsorb, nHealMin, nHealMax
	end

	-- health
	bShow = nHealthMaxAbsorb ~= plate.nHealthMaxMax
	if bShow then		-- shield and absorb are progress fills
		plate.health:SetMax(nHealthMaxAbsorb)		-- set max before progress or control breaks
		plate.nHealthMaxMax = nHealthMaxAbsorb
	end
	if nHealth ~= plate.nHealth or bShow then
		plate.health:SetProgress(nHealth, nHealthMaxAbsorb * 4)
		plate.nHealth = nHealth
	end
	if bShowHealth ~= plate.bShowHealth then
		plate.health:Show(bShowHealth)
		plate.bShowHealth = bShowHealth
	end

	-- shield
	bShow = nShieldMax ~= plate.nShieldMax
	if bShow then		-- shield and absorb are progress fills
		plate.shield:SetMax(nShieldMax)		-- set max before progress or control breaks
		plate.nShieldMax = nShieldMax
	end
	if nShield ~= plate.nShield or bShow then
		plate.shield:SetProgress(nShield, nShieldMax * 4)
		plate.nShield = nShield
	end
	bShow = bShowHealth and nShield > 0
	if bShow ~= plate.bShowShield then
		plate.shield:Show(bShow)
		plate.bShowShield = bShow
	end
	bShow = bShow and nShield > 0 and nShield ~= nShieldMax
	if bShow ~= plate.bShieldGlow then
		plate.shield:EnableGlow(bShow)
		plate.bShieldGlow = bShow
	end

	-- absorb
	bShow = nAbsorbMax ~= plate.nAbsorbMax
	if bShow then
		plate.absorb:SetMax(nAbsorbMax)
		plate.nAbsorbMax = nAbsorbMax
	end
	if nAbsorb ~= plate.nAbsorb or bShow then
		plate.absorb:SetProgress(nAbsorb, nAbsorbMax * 4)
		plate.nAbsorb = nAbsorb
	end
	bShow = bShowHealth and nAbsorb > 0
	if bShow ~= plate.bShowAbsorb then
		plate.absorb:Show(bShow, true)
		plate.bShowAbsorb = bShow
	end
end


-- create destroy		-----------------------------------------------------------------------

function NameView:OnUnitEnter(unit)		-- Created/Detroyed, temp name for file comparison
	local player = self.unitPlayer

	if not player or not player:IsValid() then
		player = GameLib.GetPlayerUnit()
		self.unitPlayer = player
	end

	local target = self.unitTarget
	if not target or not target:IsValid() then
		target = GameLib.GetTargetUnit()
		self.unitTarget = target
	end

	local focus = self.unitFocus
	if player and (not focus or not focus:IsValid()) then
		focus = player:GetAlternateTarget()
		self.unitFocus = focus
	end
end

function NameView:OnUnitDestroyed(unit)
--self._profE = (self._profE or 0) + 1											--PROF:
	local s2p,i2p,u2p = self.arShowPlates, self.arNameplates, self.arUnit2Plate

	local idUnit = unit:GetId()
	local plate = u2p[idUnit]
	if not plate then
		return
	end

	local w,pool = plate.nameplate, self.arPlatePool
	local icount,scount,limit = #i2p, #s2p, _nPoolLimit

	plate.unit = nil
	plate.unitOrMount = nil

	local bNoSave
	if w and not w:IsValid() then				-- carbine window reference bugs, also game can freeze if SetData is ever called on invalid window
		w:SetUnit(nil)
		w, plate.nameplate = nil, nil
		bNoSave = true
	end

	if w then									-- check window
		plate.bShow = false
		w:Show(false, true)
		w:SetUnit(nil)
		if plate.rewards then					--FIXME: rewards isnt set sometimes even though w is, mystery
			plate.rewards:SetData(nil)			-- release RewardIcons table tied to unit
		end

		limit = limit - (icount - scount)		-- constrain by inverse of plates to visible
		if limit < _nPoolMinLimit then limit = _nPoolMinLimit end
		pool = self.arWindowPool
	end

	local count = #pool					-- check pool
	if count < limit and not bNoSave then
		pool[count + 1] = plate
	elseif w then
		w:Destroy()
--self._profWd = (self._profWd or 0) + 1											--PROF:
--self._profPd = (self._profPd or 0) + 1											--PROF:
--	else																			--PROF:
--self._profPd = (self._profPd or 0) + 1											--PROF:
	end

	u2p[idUnit] = nil					-- remove from indexes

	for i = 1,icount do
		if i2p[i] == plate then
			i2p[i] = i2p[icount]; i2p[icount] = nil
			break
		end
	end

	for i = 1,scount do
		if s2p[i] == plate then
			s2p[i] = s2p[scount]; s2p[scount] = nil
--self._profCr = (self._profCr or 0) + 1										--PROF:
			break
		end
	end
end

function NameView:OnUnitCreated(unit)
--self._profE = (self._profE or 0) + 1											--PROF:
	local t = self.arPreloadUnits
	if t then
		t[#t + 1] = unit
		return
	end

	local idUnit = unit and unit:IsValid() and unit:GetId()
	if not idUnit or not unit:ShouldShowNamePlate() then
		return
	end
	local unitType = unit:GetType()
	if unitType == "Collectible" or unitType == "PinataLoot" then	-- never have nameplates
		return
	end
	--Print("create: " .. tostring(unit:GetName()))								--PROF:
	--	.. " age:" .. plate.nAge .. " now:" .. self.nAge)						--PROF:

	local i2p,u2p,pool = self.arNameplates, self.arUnit2Plate, self.arPlatePool

	local plate = u2p[idUnit]
	if plate then
--Print("has")																	--PROF:
		return
	end

	local count = #pool
	if count > 0 then												-- check plate pool
		plate, pool[count] = pool[count], nil
	else
		plate = { }
--self._profPc = (self._profPc or 0) + 1										--PROF:
	end

	-- set plate
	local player = self.unitPlayer
	if not player or not player:IsValid() then
		player = GameLib.GetPlayerUnit()
		self.unitPlayer = player
	end

	self:InitPlate(plate, idUnit, unit, unitType)

	i2p[#i2p + 1] = plate
	u2p[idUnit] = plate

	plate.bRequestInit = true			-- may never happen
	plate.bRequestRewards = true		-- needs just one independant update if testing gets far enough
end

function NameView:InitPlate(plate, idUnit, unit, unitType)

	if not self.tPlateTemplate then		-- make static template with base values
		self.tPlateTemplate = { }
		self:InitPlate(self.tPlateTemplate, unit:GetId(), unit, "Player")
	end

	local guild = self.tPlayerGuild
	--local guild, warparty = self.tPlayerGuild, self.tPlayerWarparty

	plate.idUnit				= idUnit
	plate.unitOwner				= unit
	plate.sUnitType				= unitType
	plate.bSpeechBubble 		= false
	plate.bGibbed				= false
	plate.bUsePvpMarker			= MatchingGameLib.IsInPvpGame() and (unitType == "Player" or unitType == "Esper Pet" or unitType == "Pet")
	plate.bIsGuildMember 		= guild and guild:IsUnitMember(unit) or false
	plate.bInCombat				= false
	--plate.bIsWarPartyMember	= warparty and warparty:IsUnitMember(unit) or false		-- not used
		-- re-set in event checks
	plate.bIsTarget 			= GameLib.GetTargetUnit() == unit
	plate.bIsCluster 			= plate.bIsTarget										-- in "target group". single always counts as cluster of one, even if not in a game cluster

	-- dependant on window state:   so leave as is or default nil
		-- re-set in slow 'visibility' checks
	plate.nAge					= 0
	plate.bIsMounted			= false
	plate.unitOrMount			= unit
	plate.bShowCast				= false													-- timer compiled settigns, as is asked every frame
	plate.bShowVuln				= false													-- timer compiled settigns, as is asked every frame
	--plate.bShow				= nil
	--plate.eDisposition		= nil
	--plate.bIsScaled			= nil													-- for color check, is set intermittently
	--plate.bShowHealthBar		= nil													-- whole health bar
	--plate.crText				= nil
	--plate.crLevel				= nil
	--plate.bShowCertainDeath	= nil
	--plate.bShowTargetSelection = nil
	--plate.bShowMarkerArrow	= nil
	--plate.bShowQuestRewards	= nil
	--plate.bIsRewardQuest		= nil
	--plate.bIsRewardContract	= nil
	--plate.bIsRewardEvent		= nil
	--plate.bRewardIcons		= nil
	--plate.bRequestLayout		= nil													-- needs nameplate layout
	--plate.bRequestInit		= nil													-- needs finish initialization

		-- re-set in fast checks
	--plate.bShowName			= nil
	--plate.sName				= nil
	--plate.bShowGuild			= nil
	--plate.sGuild				= nil
	--plate.bShowCastBar		= nil
	--plate.sCastName			= nil
	--plate.nCastDuration		= nil
	--plate.nCastElapsed		= nil
	--plate.bShowVulnBar		= nil
	--plate.nVulnerableTimeMax	= nil								-- total time
	--plate.eHealthTintType		= nil								-- background color of actual health
	--plate.bShieldGlow			= nil
	--plate.nHealth				= nil
	--plate.nHealthMax			= nil
	--plate.nShield				= nil
	--plate.nShieldMax			= nil
	--plate.nAbsorb				= nil
	--plate.nAbsorbMax			= nil
	--plate.nTotalMax			= nil
	--plate.bShowHealth			= nil
	--plate.bShowShield			= nil
	--plate.bShowAbsorb			= nil
end

function NameView:DemotePlate(plate)
	local w,pool = plate.nameplate, self.arPlatePool

	local count = #pool
	local p = pool[count]
	if not w or not p then				-- no clean plates, leave it promoted
		return
	end
	--Print("demote " .. plate.unitOwner:GetName())									--PROF:

	local wpool,ptemplate = self.arWindowPool, self.tPlateTemplate
	local wcount = #wpool

	w:SetUnit(nil)

	if wcount < _nPoolMinLimit then								-- save window
		for k,v in next, plate do
			if ptemplate[k] == nil then p[k] = v; plate[k] = nil end
		end
		pool[count], wpool[wcount + 1] = nil, p						-- change pool
	else
		for k,_ in next, plate do
			if ptemplate[k] == nil then plate[k] = nil end
		end
		w:Destroy()
--self._profWd = (self._profWd or 0) + 1											--PROF:
	end

	plate.bRequestInit = true
end

function NameView:PromotePlate(plate)
	local w,wpool = plate.nameplate, self.arWindowPool
	--Print("promote " .. plate.unitOwner:GetName())								--PROF:
	if w then									-- curently not ok to call this if already promoted
		return									--   check added to let RequestInit flag re-init certain state
	end

	local count = #wpool						-- check pool
	if count > 0 then
		local p = wpool[count]					-- window pool
		wpool[count] = nil

		local pool = self.arPlatePool			-- plate pool
		count = #pool
		if count < _nPoolLimit then
			pool[count + 1] = p
--		else																		--PROF:
--self._profPd = (self._profPd or 0) + 1											--PROF:
		end

		for k,v in next, p do				-- transfer plate data
			if plate[k] == nil then plate[k] = v; p[k] = nil end
		end
		w = plate.nameplate
	end

	if not w or not w:IsValid() then
		w = Apollo.LoadForm(self.xmlDoc, "Nameplate", "InWorldHudStratum", self)
--self._profWc = (self._profWc or 0) + 1

		plate.nameplate = w

		plate.healthBar = w:FindChild("HealthBar")
		plate.castBar = w:FindChild("CastBar")
		plate.vulnBar = w:FindChild("VulnBar")
		plate.level = w:FindChild("Level")
		plate.guild = w:FindChild("Guild")
		plate.name = w:FindChild("Name")
		plate.certainDeath = w:FindChild("CertainDeath")
		plate.scalingMark = w:FindChild("ScalingMark")
		plate.rewards = w:FindChild("Rewards")
		plate.shield = w:FindChild("Shield")
		plate.absorb = w:FindChild("Absorb")
		plate.health = w:FindChild("Health")
		plate.healAbsorb = w:FindChild("HealAbsorb")
		plate.healMin = w:FindChild("HealMin")
		plate.healMax = w:FindChild("HealMax")
		plate.healthLabel = w:FindChild("HealthLabel")
		plate.castLabel = w:FindChild("CastLabel")
		plate.castFill = w:FindChild("CastFill")
		plate.vulnFill = w:FindChild("VulnFill")
		plate.targetArrow = w:FindChild("TargetArrow")			-- arrow on top
		plate.targetMarker = w:FindChild("TargetMarker")		-- highlight and arrows sides of health bar
	end

	if not w:IsValid() or not plate.certainDeath then			--NOTE: work around for C side race condition
		local ptemplate = self.tPlateTemplate					--  by the Lua code, this state should never happen but does.
		for k,_ in next, plate do								--  only happens for breadth of one frame for only one unit, and is excedingly rare
			if ptemplate[k] == nil then plate[k] = nil end		--IMPORTANT: make sure this stays in sync with demote code
		end
		w:SetUnit(nil)
		w:Destroy()
		plate.nameplate = nil
		return													-- caller will attempt to promote again next frame
	end

	-- set window
	--plate.bShow = nil
	w:SetData(plate)
	w:SetUnit(plate.unitOrMount, 1)

	-- one time fix-ups

	-- only happen once, upon travel or enter instance all are destroyed and recreated
	if plate.certainDeath then				--FIXME: carbine side race condition issue with removing values in table
		plate.certainDeath:SetSprite(plate.bUsePvpMarker and _sPvpTargetSprite or _sHighLevelSprite)
	end
end

function NameView:UpdateRewards(plate)
	local o,unit = self.userSettings, plate.unitOwner

	local GetUnitRewardIconsForm = RewardIcons and RewardIcons.GetUnitRewardIconsForm
	if not GetUnitRewardIconsForm then
		return
	end
--self._profRI = (self._profRI or 0) + 1											--PROF:

	local w = plate.rewards or self.rewardsDummy
	GetUnitRewardIconsForm(w, unit, {
		bVert = false,
		bHideQuests = not o.bShowIconQuests,				-- covers contacts
		bHideChallenges = not o.bShowIconChallenges,
		bHideMissions = not o.bShowIconMissions,			-- covers paths
		bHidePublicEvents = not o.bShowIconPublicEvents,
		bHideRivals = not o.bShowIconRivals,
		bHideFriends = not o.bShowIconFriends,
	})

	local rewards = w:GetData()

	plate.bRewardIcons = rewards and rewards.nIcons and rewards.nIcons > 0

	plate.bIsRewardQuest, plate.bIsRewardContract, plate.bIsRewardEvent = nil, nil, nil
	if rewards and rewards.tRewardInfo then
		for _,v in pairs(rewards.tRewardInfo) do
			local eType = v.eType
			if eType == _eRewardInfoType.Quest then
				plate.bIsRewardQuest = true
			end
			if eType == _eRewardInfoType.Contract then
				plate.bIsRewardContract = true
			end
			if eType == _eRewardInfoType.PublicEvent then
				plate.bIsRewardEvent = true
			end
		end
	end

	plate.bRequestLayout = true
end

function NameView:UpdateLayout(plate)
--Print("UpdateLayout "..plate.nAge)											--PROF:
	local w,wr,ws,wg = plate.nameplate, plate.rewards, plate.scalingMark, plate.guild

	-- reward width
	local nRewardsWidth = plate.bShowQuestRewards and wr:GetWidth() or 0
	local nScaledWidth = plate.bIsScaled and ws:GetWidth() or 0

	-- arrange virt
	local left, top, right, bottom = self.nPlateLeft, self.nPlateTop, self.nPlateRight, self.nPlateBottom
	if not plate.bShowGuild then
		top = top + self.nGuildHeight
	end

	-- name
	local nHalfWidth
	if plate.bShowName or plate.bShowGuild then
		local name = plate.bShowName and Apollo.GetTextWidth(_sFont, plate.sName) or 0
		local guild = plate.bShowGuild and Apollo.GetTextWidth(_sGuildFont, plate.sGuild) or 0

		local width = name > guild and name or guild
		if width < self.nHealthRight then
			width = self.nHealthRight
		end
		nHalfWidth = ((width) + 1) / 2

		width = nHalfWidth + 17

		left, right = -width, width
	end

	-- scale effect
	local nScale = plate.nScale
	if nScale and nScale ~= 1 then
		local n = _floor(left * nScale	+ 0.5)
		right = right - (left - n)
		left = n
		n = _floor(top * nScale + 0.5)
		bottom = bottom - (top - n)
		top = n
	end

	w:SetAnchorOffsets(left, top, right, bottom)

	-- reward layout
	if plate.bShowQuestRewards then
		local width = nHalfWidth or 0
		local _, top, _, bottom = wr:GetAnchorOffsets()
		wr:SetAnchorOffsets(width, top, width + nRewardsWidth, bottom)
	end
	if plate.bIsScaled then
		local width = nHalfWidth + nRewardsWidth
		local _, top, _, bottom = ws:GetAnchorOffsets()
		ws:SetAnchorOffsets(width, top, width + nScaledWidth, bottom)
	end
end

function NameView:DrawName(plate)
	local o,unit = self.userSettings, plate.unitOwner

	if _sFont ~= plate.sNameFont then
		plate.name:SetFont(_sFont)
		plate.sNameFont = _sFont
		plate.bRequestLayout = true
	end

	local bIsTarget,bFirst,bLast = plate.bIsTarget, o.bShowFirst, o.bShowLast
	local bFull = (o.bShowFullNpc and plate.sUnitType ~= "Player")
		or (o.bShowTargetFull and bIsTarget)

	local bShow = (o.bShowName and (bFirst or bLast or bFull))
		or (o.bShowTargetName and bIsTarget)

	local sName, bTitle
	if bShow then
		bTitle = o.bShowAffiliation and (o.bShowTitle or bIsTarget)
		sName = bTitle and unit:GetTitleOrName() or unit:GetName()
		bShow = (sName and sName ~= "") and true or false
	end

	if bShow ~= plate.bShowName then
		plate.name:Show(bShow)
		plate.bShowName = bShow
		plate.bRequestLayout = true
	end

	if not bShow then
		return
	end

	if not bFull and (not bFirst or not bLast) then
		local name = not bTitle and sName or unit:GetName()

		local first, last = name:match("([^ ]+) *([^ ]*)")
		if bLast and last and last ~= "" then
			first = last
		end

		if bTitle then
			local s1, s2 = sName:match("(.*)"..name.."(.*)")
			first = (s1 or "") .. first .. (s2 or "")
		end

		sName = first
	end

	if sName ~= plate.sName then
		plate.name:SetText(sName)
		plate.sName = sName
		plate.bRequestLayout = true
	end
end

function NameView:DrawGuild(plate)
	local o,unit = self.userSettings, plate.unitOwner

	local bShow = plate.bIsTarget and o.bShowTargetAffiliation or (o.bShowAffiliation and o.bShowGuild)

	local sGuild
	if bShow then
		sGuild = unit:GetAffiliationName()
		bShow = (sGuild and sGuild ~= "") and true or false
	end

	if bShow ~= plate.bShowGuild then
		plate.guild:Show(bShow)
		plate.bShowGuild = bShow
		plate.bRequestLayout = true
	end

	if bShow then
		if unit:GetType() == "Player" then
			sGuild = String_GetWeaselString(_LNameplates_GuildDisplay, sGuild)
		end

		if sGuild ~= plate.sGuild then
			plate.guild:SetTextRaw(sGuild)
			plate.sGuild = sGuild
			plate.bRequestLayout = true
		end
	end
end

function NameView:DrawRewards(plate)
	local o = self.userSettings

	local bShow = plate.bIsTarget and o.bShowTargetRewards or o.bShowRewards

	if bShow then
		bShow = plate.bRewardIcons and true or false
	end

	if bShow ~= plate.bShowQuestRewards then
		plate.rewards:Show(bShow)
		plate.bShowQuestRewards = bShow
		plate.bRequestLayout = true
	end
end

function NameView:DrawLevel(plate)
	local unit = plate.unitOwner

	local sLevel = unit:GetLevel() or "-"
	if sLevel ~= plate.sLevel then
		plate.level:SetText(sLevel)
		plate.sLevel = sLevel
	end
end

function NameView:DrawHealthBar(plate, bInstant)
	local o,unit = self.userSettings, plate.unitOwner
	local bIsTarget = plate.bIsTarget

	local bShow = not unit:IsDead()
	if bShow then
		local nHealthMax = unit:GetMaxHealth()
		bShow = nHealthMax ~= nil
		if bShow then
			if bIsTarget then
				bShow = o.bShowTargetHealth
			else
				bShow = o.bShowHealth or (o.bShowHealthDamaged and (unit:GetHealth() ~= nHealthMax))
			end
		end
	end

	if bShow ~= plate.bShowHealthBar then
		plate.healthBar:Show(bShow, bInstant)
		plate.bShowHealthBar = bShow
	end

	bShow = bIsTarget and o.bShowTargetText or not plate.bIsTarget and o.bShowText
	if bShow ~= plate.bShowText then
		plate.healthLabel:Show(bShow, true)
		plate.bShowText = bShow
	end

	bShow = bIsTarget and o.bShowTargetMaxHealthText or not plate.bIsTarget and o.bShowMaxHealthText
	if bShow ~= plate.bShowMaxHealthText then
		plate.bNewTextOption = true
		plate.bShowMaxHealthText = bShow
	end

	bShow = bIsTarget and o.bShowTargetShieldText or not plate.bIsTarget and o.bShowShieldText
	if bShow ~= plate.bShowShieldText then
		plate.bNewTextOption = true
		plate.bShowShieldText = bShow
	end
end

function NameView:DrawTargeting(plate, bInstant)
	local o = self.userSettings

	local bShow = o.bShowTargetSelection and plate.bIsTarget
	local bShowArrow = bShow

	if bShow then
		bShow = plate.bShowHealthBar
		bShowArrow = not bShow
	end

	if bShow ~= plate.bShowTargetSelection then
		plate.targetMarker:Show(bShow, bInstant)
		plate.bShowTargetSelection = bShow
	end

	if bShowArrow ~= plate.bShowTargetArrow then
		plate.targetArrow:Show(bShowArrow, not bShowArrow or bInstant)
		plate.bShowTargetArrow = bShowArrow
	end
end


-- visibility timer		-----------------------------------------------------------------------

function NameView:OnUpdateTimer()
	self.nAge = self.nAge + 1
	--local nTime = GameLib.GetTickCount()												--PROF:
	--self._frameCount = (self._frameCount or 0) + 1

	-- animation

	if (self.nAge % _nAniSpeed) ~= 0 then
		local s2a,player = self.arAniPlates, self.unitPlayer
		if player and self.bIsPlayerMove and next(s2a) then
			self.tPlayerPos = player:GetPosition() or self.tPlayerPos
		end
		for k,v in next, s2a do
			self:UpdateAnimation(k)
			s2a[k] = nil
		end
		return
	end

	-- visibility

	self:UpdateVisibilityCommon()

--for i = 1,10 do																		--PROF:
	for _,v in next, self.arNameplates do
		self:UpdateVisibility(v)
	end
	self.bRequestRewards = false
--end																					--PROF:
	--local s2p,i2p,u2p,pool,wpool = self.arShowPlates, self.arNameplates,				--PROF:
	--self.arUnit2Plate, self.arPlatePool, self.arWindowPool							--PROF:
	--if self.bIsTarget and self.nAge % 5 == 1 then										--PROF:
	--	Print( "" .. #i2p .. " s:" .. #s2p												--PROF:
	--		.. " p: " .. #pool .. "." .. #wpool											--PROF:
	--		.. " w: " .. (self._profWc or 0) .. "." .. (self._profWd or 0)				--PROF:
	--		.. " p: " .. (self._profPc or 0) .. "." .. (self._profPd or 0))				--PROF:
	--end																				--PROF:
	--if self._frameCount % 30 == 1 then												--PROF
	--	Print((self._frameCount or 0) .. ": " .. (GameLib.GetTickCount() - nTime))		--PROF:
	--end																				--PROF:
end

function NameView:UpdateVisibilityCommon()
	local o,player = self.userSettings, self.unitPlayer

	local u = self.tPlayerPos
	local p = player and player:GetPosition() or u
	self.bIsPlayerMove = ((p and u) ~= nil)								-- if missing values, then dont let bMove stay latched
		and (p.x ~= u.x or p.y ~= u.y or p.z ~= u.z)
	self.tPlayerPos = p

	local nFovY = Apollo.GetConsoleVariable("camera.FovY") or 60

	local nUiScale = Apollo.GetConsoleVariable("ui.Scale")

	if nUiScale ~= self.nUiScale or nFovY ~= self.nFovY then
		self.nFovY = nFovY
		self.nUiScale = nUiScale

		local t = Apollo.GetDisplaySize()
		self.nUiWidth,self.nUiHeight = t.nRawWidth, t.nRawHeight

		self:OnSettingFontChanged()
	end
end

function NameView:UpdateVisibility(plate)
	local o,w,unit,player = self.userSettings, plate.nameplate, plate.unitOwner, self.unitPlayer

	if not player or not unit:IsValid() then		-- world or zone unit in transition
		return
	end

	-- up-front transitions

	local eDisposition = unit:GetDispositionTo(player)
	if eDisposition ~= plate.eDisposition then
		plate.eDisposition = eDisposition
	end

	local bIsMounted = unit:IsMounted()									--FIXME: might be able to remove mounted swap, via latest patch
	local unitOrMount = bIsMounted and unit:GetUnitMount() or unit		--  would need lots of testing with GetUnitScreenPosition and IsOccluded
	if unitOrMount ~= plate.unitOrMount then							-- swap for horse or not
		if w then														-- else set by window init below if shown
			w:SetUnit(unitOrMount)
		end
		plate.unitOrMount = unitOrMount
	end
	if bIsMounted ~= plate.bIsMounted then								-- swap for horse or not
		plate.bIsMounted = bIsMounted
	end

	-- check options

	local bShow = not plate.bGibbed and (o.bShowDuringSpeech or not plate.bSpeechBubble)
	if bShow then
		if w then	-- switch occlusion to sync window position is updates
			bShow = (not o.bUseOcclusion or not w:IsOccluded()) and w:IsOnScreen()
		else
			bShow = (not o.bUseOcclusion or not unitOrMount:IsOccluded())
				and (GameLib.GetUnitScreenPosition(unitOrMount) or {}).bOnScreen  -- returns nil once in a blue moon, will almsot never create table
		end
	end

	if self.bRequestRewards then
		plate.bRequestRewards = true
	end
	if bShow and plate.bRequestRewards then									-- needed for proper objectives check
		self:UpdateRewards(plate)											-- very expensive
		plate.bRequestRewards = nil
	end
	--if bShow and (self.bRequestRewards or plate.bRequestRewards) then		-- needed for proper objectives check
	--	self:UpdateRewards(plate)											-- very expensive
	--	plate.bRequestRewards = nil
	--end

	--if not bShow or plate.bIsCluster then									--NOTE: for questing is less confusing when you hit a neutral pack at edge of range
	if not bShow or plate.bIsTarget then

	elseif unit == player then
		bShow = o.bShowMyPlate and not unit:IsDead()

	elseif not unit:ShouldShowNamePlate() then								-- im not sure this is causing the right behaviour
		bShow = false

	elseif o.bShowObjectives and (plate.bIsRewardQuest
		or plate.bIsRewardContract or plate.bIsRewardEvent) then

	elseif (o.bShowInCombat or not self.bPlayerInCombat) then
		bShow = false
		if (o.bShowFriendlyPlayer and eDisposition == _eDisposition.Friendly 	-- simple ones
				and (plate.sUnitType == "Player" or o.bShowFriendly))
			or (o.bShowHostile and eDisposition == _eDisposition.Hostile
				and (o.bShowEnemysUnits or (plate.sUnitType ~= "Pet" and plate.sUnitType ~= "Esper Pet")))
			or (o.bShowNeutral and eDisposition == _eDisposition.Neutral)
			or (o.bShowOrganization and plate.bIsGuildMember)
			or (o.bShowGroup and unit:IsInYourGroup()) then						-- minimum, defaults never called
			bShow = true

		elseif o.bShowVendor or o.bShowTaxi or o.bShowObjectives then			-- ones that require expensive State call
			local state = unit:GetActivationState()								-- nested multi-table copy each call, called for any non-objective non-player
			if (o.bShowVendor and state.Vendor)									--  only "friendly units" option would preempt which is off by default
				or (o.bShowTaxi and (state.FlightPathSettler or state.FlightPath or state.FlightPathNew))
				or (o.bShowObjectives and (state.QuestReward or state.QuestNew
					or state.QuestNewMain or state.QuestReceiving or state.TalkTo)) then
				bShow = true
			end
		end
	else
		bShow = false
	end

	-- check distance

	if bShow then														-- very expensive, 3x more than original Objectives visi test
		local u,p = unit:GetPosition(), self.tPlayerPos				--  new table makes dist check necessarly dead last check, should be near first check
		bShow = (u and p) ~= nil										--  300 units means 600 non-empty hash based tables, just here
		if bShow then													--	test cratered here as is where would often push the GC against the wall
			local r = plate.bIsCluster and _nTargetMax or o.nMaxRange	-- cluster always includes direct target here
			local x,y,z = u.x - p.x, u.y - p.y, u.z - p.z
			y = x * x + y * y + z * z
			bShow = y < (r * r)
			plate.nDistance = y
			u, plate.tPos = plate.tPos, u
			plate.bIsMove = u and (u.x ~= x or u.y ~= y or u.z ~= z)
		end
	end

	-- configure visibility

	if bShow and plate.bRequestInit then
		self:PromotePlate(plate)
		w = plate.nameplate
	end
	if not bShow and w and self.nAge - plate.nAge > _nReclamationAge then
		self:DemotePlate(plate)
		w = plate.nameplate
	end
	if not w then
		return
	end

	if bShow ~= plate.bShow then
		w:Show(bShow, plate.bIsCluster or (not plate.bRequestInit and self.bPlayerInCombat)) -- only fade in or out plates if not target, or is new and not in combat
		plate.bShow = bShow

		local s2p = self.arShowPlates
		local count = #s2p
		if bShow then
			s2p[count + 1] = plate
		else
			for i = count,1,-1 do
				if s2p[i] == plate then
					s2p[i] = s2p[count]; s2p[count] = nil; break
				end
			end
		end
	end

	-- configure display

	if bShow then
		plate.nAge = self.nAge
		--self:DrawHealth(plate)		-- fast only

		self:DrawMisc(plate)			-- slow only
		--self:DrawAnimation(plate)		-- slow only.  full version handled in misc
		self:DrawTextColor(plate)		-- slow only.  must be after misc

		self:DrawHealthBar(plate)		-- create, slow, target change
		self:DrawTargeting(plate)		-- create, slow, target change. must be after health
		self:DrawLevel(plate)			-- create, event.  client bugs require checking this every time for NPCs rather than relying on the event, very rare but ther are issues

		if plate.bRequestInit then
			self:DrawName(plate)		-- create, event, target change
			self:DrawGuild(plate)		-- create, event, target change
			plate.bRequestInit = nil
			plate.bRequestRewards = true	-- don't populate now, wait for next test. unit may be oscilating
		end

		self:DrawRewards(plate)			-- create, slow

		if plate.bRequestLayout then
			self:UpdateLayout(plate)
			plate.bRequestLayout = false
		end
	end
end

-- configure misc items
--  sets: plate.bIsScaled
function NameView:DrawMisc(plate)
	local o,w,unit = self.userSettings, plate.nameplate, plate.unitOwner

	local bIsTarget = plate.bIsTarget

	-- certain death or pvp
	local bShow = plate.eDisposition ~= _eDisposition.Friendly
	if bShow then
		bShow = (plate.bUsePvpMarker and (o.bShowMarker or bIsTarget and o.bShowTargetMarker)
			or (o.bShowCertainDeath and unit:IsCertainDeath())) and true or false
	end

	if bShow then
		bShow = unit:ShouldShowNamePlate()						-- flagged to display
			and unit:GetHealth()								-- has health value
			and not unit:IsDead()								-- is not dead
	end
	bShow = bShow and true or false
	if bShow ~= plate.bShowCertainDeath then
		plate.certainDeath:Show(bShow)
		plate.bShowCertainDeath = bShow
	end

	-- cast bar
	plate.bShowCast = o.bShowCastBar							-- timer compiled settings for every shown plate on every frame
		or plate.bIsTarget and o.bShowTargetCastBar

	-- vuln bar
	plate.bShowVuln = (o.bShowVulnBar
		or plate.bIsTarget and o.bShowTargetVulnBar)
		and (not o.bShowHealthNever
		or plate.bIsTarget and o.bShowTargetHealth)

	-- level scaled icon
	local bIsScaled = unit:IsScaled()
	if bIsScaled ~= plate.bIsScaled then
		plate.scalingMark:Show(bIsScaled)
		plate.bIsScaled = bIsScaled								-- only place bIsScaled is set, used by GetColorNameplate
		plate.bRequestLayout = true
	end

	-- perspective

	bShow = (o.bPerspectiveScale or o.bPerspectiveOpacity) and (o.bPerspectiveCombat or not self.bPlayerInCombat)
		and (o.bPerspectiveCombatants or not plate.bInCombat) and (o.bTargetPerspective or not plate.bIsCluster)

	self:DrawAnimation(plate, bShow)
end

-- configure misc items
--  sets: plate.bIsScaled
function NameView:DrawAnimation(plate, bShow, bIsTween)
	local o,unit = self.userSettings, plate.unitOwner

	local nScale, nOpacity = 1, 1

	if bShow then
		local n = (plate.nDistance or 0) / (_nRangeMaxSq * _nVanishCurve)		-- dist^2 / valishing point curve. shallow close, steep far

		if n > _nNearClip then
			n = 1 - (n < _nFarClip and n or _nFarClip) * _nVanishTrim
			if o.bPerspectiveScale then
				nScale = _floor(n * 100 + 0.5) / 100										-- .01 steps
			end
			if not bIsTween then
				if o.bPerspectiveOpacity then
					nOpacity = ((_floor(n * 24 + 0.5) / 24) ^ _nOpacityCurve) * _nOpacityTrim	-- .05 steps
				elseif nScale ~= 1 then
					nOpacity = nScale > .70 and .90 or .86		-- tiny bit compensate outline mangle and bottom clip region artifacts
				end
			end
		end
	end

	--if plate.bIsTarget and (nScale ~= plate.nScale or nOpacity ~= plate.nOpacity) then					--PROF:
	--	Print(tostring(plate.nAge or 0) .. ": " .. _floor(_sqrt(plate.nDistance or 0)) .. " = "		--PROF:
	--		.. _floor((1-nScale) * 1000) .. " " .. _floor((1-nOpacity) * 1000) ) end				--PROF:

	local nScalePlate = plate.nScale
	if nScale ~= nScalePlate then
		if bShow and nScalePlate and nScale < .99 and not plate.bRequestInit and not bIsTween then
			self.arAniPlates[plate] = true
		end
		--if plate.bIsTarget then																	--PROF:
		--	Print(tostring(nScale) .. (bIsTween and " ." or ""))									--PROF:
		--end																						--PROF:
		plate.nameplate:SetScale(nScale, false)
		plate.nScale = nScale
		plate.bRequestLayout = true
	end

	if not bIsTween and nOpacity ~= plate.nOpacity then
		plate.nameplate:Show(false, false)				-- try to start competing animation fading down at same time,
		plate.nameplate:Show(true, true)				--  because of the opacity pulse bug
		plate.nameplate:SetOpacity(nOpacity == 1 and 1 or (nOpacity ^ .975 * .975), 2000)	-- for same reasons, extra w/ invert curve a bit to help lower update frequency
		plate.nOpacity = nOpacity
	end
end

function NameView:UpdateAnimation(plate)
	local o,unit,player = self.userSettings, plate.unitOwner, self.unitPlayer

	if not player or not unit:IsValid() then		-- world or zone unit in transition
		return
	end

	local bShow = plate.bShow

	-- check distance

	if bShow then
		--local u,p = unit:GetPosition(), self.tPlayerPos
		local u,p = (plate.bIsMove and unit:GetPosition() or plate.tPos), self.tPlayerPos
		if (u and p) ~= nil then
			local x,y,z = u.x - p.x, u.y - p.y, u.z - p.z
			plate.nDistance = x * x + y * y + z * z
		end
	end

	-- configure display

	if bShow then
		plate.nAge = self.nAge

		self:DrawAnimation(plate, true, true)			-- slow only

		if plate.bRequestLayout then
			self:UpdateLayout(plate)
			plate.bRequestLayout = false
		end
	end
end

-- configure text colors
function NameView:DrawTextColor(plate)
	local unit,player = plate.unitOwner, self.unitPlayer
	local crText, crLevel

	local bIsScaled = plate.bIsScaled

	-- level

	local bIsDead = unit:IsDead()

	if bIsDead then
		crLevel = _crDead
	elseif bIsScaled then
		crLevel = _crScaling
	else
		local n = (player:GetLevelDifferential(unit) or 0) + _nConOffset
		crLevel = _arConColor[n < _nConMin and _nConMin or n] or _arConColor[_nConMax]
	end

	-- color and group size

	local unitType = plate.sUnitType

	if bIsDead then
		crText = _crDead

	elseif unitType == "Player" or unitType == "Pet" or unitType == "Esper Pet" then
		local owner = (unitType == "Pet" or unitType == "Esper Pet") and unit:GetUnitOwner() or unit
		local bOwnerFlag = owner:IsPvpFlagged() and true or false

		if plate.eDisposition == _eDisposition.Friendly or owner:IsThePlayer() then
			crText = bOwnerFlag and _crAlly
				or owner:IsInYourGroup() and _crGroup
				or plate.bIsGuildMember and _crGuild or _crAllyUnflag
		else
			local bPlayerFlagged = GameLib.IsPvpFlagged() and true or false
			crText = bOwnerFlag ~= bPlayerFlagged and _crEnemyMismatch
				or bOwnerFlag and _crEnemy or _crEnemyUnflag
		end

	elseif unitType ~= "Player" and unit:IsTagged() and not unit:IsTaggedByMe() and not unit:IsSoftKill() then
		crText = _crDefaultTagged
	else													-- npc
		crText = _tTextColors[plate.eDisposition]
	end

	if crText ~= plate.crText then
		plate.name:SetTextColor(crText)
		plate.guild:SetTextColor(crText)
		plate.crText = crText
	end
	if crLevel ~= plate.crLevel then
		plate.level:SetTextColor(crLevel)
		plate.crLevel = crLevel
	end
end


-- other system events	-----------------------------------------------------------------------

function NameView:OnRequestRewardsUpdate()
--self._profRUA = (self._profRUA or 0) + 1									--PROF:
	self.bRequestRewards = true
end

function NameView:OnUnitTextBubbleToggled(unit, strText, nRange)
--self._profE = (self._profE or 0) + 1										--PROF:
	local plate = self.arUnit2Plate[unit:GetId()]
	if plate then
		plate.bSpeechBubble = strText and strText ~= ""
		self:UpdateVisibility(plate)
	end
end

function NameView:OnEnteredCombat(unit, bInCombat)
--self._profE = (self._profE or 0) + 1										--PROF:
	if unit == self.unitPlayer then
		self.bPlayerInCombat = bInCombat
	end
	local plate = self.arUnit2Plate[unit:GetId()]
	if plate then
		plate.bInCombat = bInCombat
		self:UpdateVisibility(plate)
	end
end

function NameView:OnUnitGibbed(unit)
--self._profE = (self._profE or 0) + 1										--PROF:
	local plate = self.arUnit2Plate[unit:GetId()]
	if plate then
		plate.bGibbed = true
		self:UpdateVisibility(plate)
	end
end

function NameView:OnUnitNameChanged(unit, strNewName)
--self._profE = (self._profE or 0) + 1										--PROF:
	local plate = self.arUnit2Plate[unit:GetId()]
	if plate and plate.nameplate then
		self:DrawName(plate)
	end
end

function NameView:OnUnitTitleChanged(unit)
--self._profE = (self._profE or 0) + 1										--PROF:
	local plate = self.arUnit2Plate[unit:GetId()]
	if plate and plate.nameplate then
		self:DrawName(plate)
	end
end

function NameView:OnPlayerTitleChanged()
--self._profE = (self._profE or 0) + 1										--PROF:
	local player = self.unitPlayer
	if not player then
		return
	end
	local plate = self.arUnit2Plate[player:GetId()]

	if plate and plate.nameplate then
		self:DrawName(plate)
	end
end

function NameView:OnUnitLevelChanged(unit)
--self._profE = (self._profE or 0) + 1										--PROF:
	local plate = self.arUnit2Plate[unit:GetId()]
	if plate and plate.nameplate then
		self:DrawLevel(plate)
	end
end

function NameView:OnGuildChange()
--self._profE = (self._profE or 0) + 1										--PROF:
	self:SetPlayerGuilds()							-- have our guilds changed

	local guild = self.tPlayerGuild					-- find guildmates
	for _,v in next, self.arNameplates do
		v.bIsGuildMember = guild and guild:IsUnitMember(v.unitOwner) or false
	end
	--local guild, warparty = self.tPlayerGuild, self.tPlayerWarparty	--NOTE: warparty color. disabled as is disabled in carbine version
	--for _,v in next, self.arNameplates do
	--	local unit = v.unitOwner
	--	plate.bIsGuildMember = guild and guild:IsUnitMember(unit) or false
	--	plate.bIsWarPartyMember = warparty and waryparty:IsUnitMember(unit) or false
	--end
end

function NameView:OnUnitGuildNameplateChanged(unit)
--self._profE = (self._profE or 0) + 1										--PROF:
	local plate = self.arUnit2Plate[unit:GetId()]
	if plate and plate.nameplate then
		self:DrawGuild(plate)
	end
end

function NameView:OnUnitMemberOfGuildChange(unit)
--self._profE = (self._profE or 0) + 1										--PROF:
	local plate = self.arUnit2Plate[unit:GetId()]
	if plate then
		local guild = self.tPlayerGuild
		plate.bIsGuildMember = guild and guild:IsUnitMember(unit) or false
		if plate.nameplate then
			self:DrawGuild(plate)
		end
	end
	--local plate = self.arUnit2Plate[unit:GetId()]						--NOTE: warparty color. disabled as is disabled in carbine version
	--if plate then
	--	local guild, warparty = self.tPlayerGuild, self.tPlayerWarparty
	--	plate.bIsGuildMember = guild and guild:IsUnitMember(unit) or false
	--	plate.bIsWarPartyMember = warparty and warparty:IsUnitMember(unit) or false
	--	self:DrawGuild(plate)
	--end
end

function NameView:OnTargetUnitChanged(unit)
--self._profE = (self._profE or 0) + 1										--PROF:
	local i2p,u2p = self.arNameplates, self.arUnit2Plate

	for _,v in next, i2p do						-- clear all
		if v.bIsCluster then					-- target is never marked without also marking cluster
			v.bIsTarget = false
			v.bIsCluster = false

			if v.nameplate then
				v.bRequestInit = true
				self:UpdateVisibility(v)
			end
		end
	end

	local plate = unit and u2p[unit:GetId()]
	if plate then
		plate.bIsTarget = true					-- target not in a "game cluster" will still be marked as cluster here
		plate.bIsCluster = true					-- unit is never returned in GetClusterUnits list

		if plate.nameplate then
			plate.bRequestInit = true
			self:UpdateVisibility(plate)
		end

		local cluster = unit:GetClusterUnits()	-- find other "target cluster" members
		if cluster and #cluster > 0 then
			for _,v in next, cluster do			-- mark other cluster units
				v = u2p[v:GetId()]
				if v then
					v.bIsCluster = true

					if v.nameplate then
						v.bRequestInit = true
						self:UpdateVisibility(v)
					end
				end
			end
		end
	end

--	self.bIsTarget = plate								--PROF:
end

function NameView:OnResolutionChanged()
	self.nUiScale = nil													-- force scale re-evaluation
	self:OnUpdateTimer()												-- immediate, so no post rez change pop
end


-----------------------------------------------------------------------------------------------
-- Nameplate Form Functions
-----------------------------------------------------------------------------------------------

--function NameView:OnNameplateNameClick(wndHandler, wndControl, eMouseButton)		--NOTE: window events were removed. no mouse-up on Carbine ignore mouse setting, so no way to stop the up event agaisnt world from immediately removing the selection
--	local w = wndHandler
--	while w and w:GetName() ~= "Nameplate" do
--		w = w:GetParent()
--	end
--	local plate = w and w:GetData()
--	if not plate then
--		return
--	end

--	local unit = plate.unitOwner
--	if eMouseButton == GameLib.CodeEnumInputMouse.Left and GameLib.GetTargetUnit() ~= unit then
--		GameLib.SetTargetUnit(unit)
--	end
--end

function NameView:OnWorldLocationOnScreen(wndHandler, wndControl, bOnScreen)
--self._profE = (self._profE or 0) + 1											--PROF:
	local plate = wndHandler and wndHandler:GetData()

	if plate then
		--plate.bOnScreen = bOnScreen
		self:UpdateVisibility(plate)
	end
end

function NameView:OnUnitOcclusionChanged(wndHandler, wndControl, bIsOccluded)
--self._profE = (self._profE or 0) + 1											--PROF:
	local plate = wndHandler and wndHandler:GetData()

	if plate then
		--plate.bIsOccluded = bIsOccluded
		self:UpdateVisibility(plate)
	end
end


-----------------------------------------------------------------------------------------------
-- OptionsForm Form Functions
-----------------------------------------------------------------------------------------------

function NameView:OnConfigure()
	self:OnOptionsForm()
end

function NameView:OnCloseWindow()
	self.wndOpts:Close()
end

function NameView:OnCancel()
	self:OnCloseWindow()
end

function NameView:OnOK(wndHandler, wndControl)
	self.configCommitting = true
	self:OnCloseWindow()
end

function NameView:OnOptionsClosed(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return
	end
	local wo,L = self.wndOpts, self.L
	if wo and wo:IsValid() then
		wo:Close()
		wo:Destroy()
		self.wndOpts = nil
		self._tOptionsTabs = nil
		self.woChoiceScroll = nil

		Event_FireGenericEvent("WindowManagementRemove", {strName=Apollo.GetString("Nameplates_Options")})
	end
end

function NameView:OnOptionsForm()
	local L = self.L

	local wo = Apollo.LoadForm(self.xmlDoc, "OptionsForm", nil, self)
	self.wndOpts = wo

	Apollo.LoadForm(self.xmlDoc, "NormalView", wo:FindChild("NormalContent"), self)
	Apollo.LoadForm(self.xmlDoc, "TargetView", wo:FindChild("TargetContent"), self)
	Apollo.LoadForm(self.xmlDoc, "AdvancedView", wo:FindChild("AdvancedContent"), self)

	self:OnPanelForm(wo)

	-- fixups
	self._tOptionsTabs = {
		NormalViewButton = "NormalContent",
		TargetViewButton = "TargetContent",
		AdvancedViewButton = "AdvancedContent"
	}
	local w = wo:FindChild(self.sOptionsTab or "NormalViewButton")
	self:OnOptionsTabCheck(nil, w)
	w:SetCheck(true)

	-- reward mission type icon
	wo:FindChild("ShowRewardTypeMission"):FindChild("Icon"):SetSprite(_arPathSprites[PlayerPathLib.GetPlayerPathType()])

	self:OnCurrentToConfig()
	wo:Invoke()

	Event_FireGenericEvent("WindowManagementAdd", {wnd = wo, strName=Apollo.GetString("Nameplates_Options"), nSaveVersion=_VersionWindow})
end

function NameView:OnCurrentToConfig()
	local o,wo = self.userSettings, self.wndOpts

	-- Generic managed controls
	self:OnCurrentToPanel(wo)

	wo:FindChild("ResetDropToggle"):SetText(o.sSettingsName)

	-- update sliders
	wo:FindChild("DrawDistanceSlider"):SetValue(o.nMaxRange)
	wo:FindChild("DrawDistanceLabel"):SetText(String_GetWeaselString(Apollo.GetString("Nameplates_DrawDistance"), o.nMaxRange))

	self:OnEnablerButtonChecked()												-- enable disable buttons
end

function NameView:OnEnablerButtonChecked()
	local o,wo = self.userSettings, self.wndOpts

	self:SetButtonEnabled(wo:FindChild("MainShowDisposition_FriendlyPlayer"), not o.bShowFriendly)		-- friendly unit
	self:SetButtonEnabled(wo:FindChild("AdvFontNameLargerer"), o.bFontLargeName)						-- health text
	self:SetButtonEnabled(wo:FindChild("AdvShowMaxHealthText"), o.bShowText)							-- health text
	self:SetButtonEnabled(wo:FindChild("AdvShowShieldText"), o.bShowText)
	self:SetButtonEnabled(wo:FindChild("TargetShowMaxHealthText"), o.bShowTargetText)
	self:SetButtonEnabled(wo:FindChild("TargetShowShieldText"), o.bShowTargetText)
	self:SetButtonEnabled(wo:FindChild("AdvShowFullNpc"), not o.bShowFirst or not o.bShowLast or not o.bShowTitle)	-- First Last Name
	self:SetButtonEnabled(wo:FindChild("AdvPerspectiveCombat"), o.bPerspectiveScale or o.bPerspectiveOpacity)		-- Perspective
	self:SetButtonEnabled(wo:FindChild("AdvPerspectiveCombatants"), o.bPerspectiveScale or o.bPerspectiveOpacity)
end

function NameView:OnSettingFontChanged()
	local o,wo = self.userSettings, self.wndOpts

	_crGuild = o.bColorLightGuild and _crGuild_Light or _crGuild_Dark
	_crEnemy = o.bColorLightEnemy and _crEnemy_Light or _crEnemy_Dark
	_tTextColors[_eDisposition.Hostile] = _crEnemy

	local n = self.nUiScale

	n = (n < 0.79 and (self.nUiWidth < 1030 or self.nUiHeight < 782) and 0.65)	-- panic modes for threshholds when the carbine client becomes unreadable
		or (n < 1.1 and (self.nUiWidth < 1020 or self.nUiHeight < 600) and 0.65) or n

	_sFont = not o.bFontLargeName and _sFont_Default
		or (n > 0.86 and (o.bFontLargerName and _sFont_Larger or _sFont_Large))
		or (n > 0.78 and (o.bFontLargerName and _sFont_Larger85 or _sFont_Large85))
		or (n >= 0.70 and (o.bFontLargerName and _sFont_Larger75 or _sFont_Large75))
		or (o.bFontLargerName and _sFont_Larger70 or _sFont_Large70)

	_nNearClip = n > .86 and _nNearClip_Default
		or n > .78 and _nNearClip_85
		or _nNearClip_75
--Print(_sFont .. " " .. _nNearClip .. " " .. self.nUiWidth .. " " .. self.nUiHeight)									--PROF:

	for _,v in next, self.arNameplates do
		v.crText = nil
		v.crLevel = nil
		if v.nameplate and v.unitOwner then
			self:DrawTextColor(v)
			self:DrawName(v)
			self:DrawGuild(v)
		end
	end

	if wo then
		self:OnEnablerButtonChecked()
	end
end

function NameView:OnSettingNameChanged()
	for _,v in next, self.arNameplates do
		if v.nameplate then
			self:DrawName(v)
		end
	end
	self:OnEnablerButtonChecked()
end

function NameView:OnSettingTitleChanged()
	for _,v in next, self.arNameplates do
		if v.nameplate then
			self:DrawName(v)
			self:DrawGuild(v)
		end
	end
	self:OnEnablerButtonChecked()
end

function NameView:OnSettingHealthChanged()
	for _,v in next, self.arNameplates do
		if v.nameplate then
			self:DrawHealthBar(v)
		end
	end
end

function NameView:OnDrawDistanceSlider(wndHandler, wndControl, fValue, fOldValue)
	local o,wo = self.userSettings, self.wndOpts

	fValue = _floor(fValue * 100) / 100					-- clip to 0.00

	wo:FindChild("DrawDistanceLabel"):SetText(String_GetWeaselString(Apollo.GetString("Nameplates_DrawDistance"), fValue))
	o.nMaxRange = fValue
end

function NameView:OnOcclusionCheck()
	local o = self.userSettings
	Apollo.SetConsoleVariable("ui.occludeNameplatePositions", o.bUseOcclusion)
end


-----------------------------------------------------------------------------------------------
-- NameView Startup
-----------------------------------------------------------------------------------------------

function NameView:OnSave(eType)
	local tSave

	if eType == GameLib.CodeEnumAddonSaveLevel.Character then
		local o = self.userSettings

		tSave = { }
	end

	if eType == GameLib.CodeEnumAddonSaveLevel.Realm then
	end

	if tSave then
		local o = self.userSettings

		for k,v in pairs(o) do
			local t = _defaultSettings[k]
			if (type(t) == "table" and t.eSaveLevel or _eSaveLevel.Character) == eType then
				tSave[k] = DeepCopy(v)
			end
		end
	end

	return tSave
end

function NameView:OnRestore(eType, tSave)
	if tSave then
		local o,_ds = self.userSettings, _defaultSettings

		local nVersion = tSave["nSettingsVersion"] or 0

		for k,v in pairs(tSave) do
			local t = _defaultSettings[k]
			if (type(t) == "table" and t.eSaveLevel or _eSaveLevel.Character) == eType then
				-- migrations
				if k == "bShowMyNameplate" and nVersion <= 0 and tSave["bShowMyPlate"] == nil then k = "bShowMyPlate" end
				if k == "bShowTitle" and nVersion <= 0 and tSave["bShowAffiliation"] == nil then k = "bShowAffiliation" end
				if k == "bShowTargetMarker" and nVersion <= 0 and tSave["bShowTargetSelection"] == nil then k = "bShowTargetSelection" end
				if k == "bShowTargetGuild" and  nVersion <= 0 and tSave["bShowTargetAffiliation"] == nil then k = "bShowTargetAffiliation" end
				if k == "bShowPremium" or k == "bShowTargetPremium" and nVersion <= 0 then v = nil end
				-- filters
				if k == "nSettingsVersion" then v = o.nSettingsVersion end
				if k == "sSettingsTooltip" then v = nil end
				o[k] = DeepCopy(v)
			end
			-- migrations
			if not tSave["bShowHealthNever"] and nVersion <= 0 then o.bShowHealthNever = not o.bShowHealth and not o.bShowHealthDamaged end
		end
	end

	if eType == GameLib.CodeEnumAddonSaveLevel.Character then
		self:InitializeWindows()
	end
end

function NameView:InitializeWindows()
	local o = self.userSettings

	self:OnCurrentToCommand()

	self.nFovY = self.nFovY or Apollo.GetConsoleVariable("camera.FovY") or 60

	self.nUiScale = self.nUiScale or Apollo.GetConsoleVariable("ui.Scale")
	local t = Apollo.GetDisplaySize()
	self.nUiWidth,self.nUiHeight = t.nRawWidth, t.nRawHeight

	Apollo.SetConsoleVariable("ui.occludeNameplatePositions", o.bUseOcclusion)

	-- re-bind any newly loaded custom reset settings to base custom settings
	self._resetSettingsCustom = self._resetSettingsCustom or _resetSettings.custom
	_resetSettings.custom = setmetatable(o.tSettingsCustom, { __index = self._resetSettingsCustom })

	-- rest needs docload
	if not self.onXmlDocLoadedCalled then
		return
	end

	o.bIsDemo = IsDemo()						-- too expensive to call 8 times per mousedown event

	local player = GameLib.GetPlayerUnit()
	self.unitPlayer = player

	-- loaded units
	self:OnSettingFontChanged()					-- update any existing units, includes OnSettingNameChanged(), OnSettingFontChanged()
	self:OnSettingHealthChanged()

	-- cached units
	local t = self.arPreloadUnits				-- add new units if any cached
	if t then
		self.arPreloadUnits = nil
--local nTime = GameLib.GetTickCount()																--PROF:
		for _,v in next, t do
			self:OnUnitCreated(v)
		end
--Print((self._frameCount or 0) .. ": init " .. #t .. " " .. (GameLib.GetTickCount() - nTime))		--PROF:
	end
end


-----------------------------------------------------------------------------------------------
-- NameView Commands
-----------------------------------------------------------------------------------------------

function NameView:OnShowMenu()				-- used by InterfaceMenuList
	self:OnCommand()
end

function NameView:OnCurrentToCommand()
	local o = self.userSettings
	print = o.bIsDebug and _print or _dummy
end

function NameView:OnCommand(sCommand, sArgs)
	local o,wo,L = self.userSettings, self.wndOpts, self.L

	sArgs = sArgs and sArgs:lower() or ""
	local sArg1, sArg2 = sArgs:match("([^ ]+) *(.-) *$")	-- '  aaa   bb cc  ' to 'aaa', 'bb cc'

	if not sArg1 or sArg1 == ""
		or sArg1 == "options"								-- toggle config window
		or sArg1 == "config" or sArg1 == "settings" then
		if wo then
			self:OnOptionsClosed()
		else
			self:OnOptionsForm()
		end
	elseif sArg1 == "optionsopen" then
		if not wo then
			self:OnOptionsForm()
		end
	elseif sArg1 == "optionsclose" then
		if wo then
			self:OnOptionsClosed()
		end

	elseif sArg1 == "reset" then
		local t = (not sArg2 or sArg2 == "") and _resetSettings.default or _resetSettings[sArg2]
		if not t then
			PostChat(_format(L["No reset settings named '%s' for %s."], sArg2, _AddonName))
			return
		end
		o = SettingsCopy(_defaultSettings); self.userSettings = o
		o.tSettingsCustom = _resetSettings.custom
		t.nSettingsVersion = t.nSettingsVersion or o.nSettingsVersion
		self:OnRestore(GameLib.CodeEnumAddonSaveLevel.Realm, t)			-- realm first because char fires init
		self:OnRestore(GameLib.CodeEnumAddonSaveLevel.Character, t)
		if wo and wo:IsValid() then
			self:OnCurrentToConfig()
		end

	elseif sArg1 == "debug" then
		o.bIsDebug = not o.bIsDebug
		self:OnCurrentToCommand()

	elseif sArg1 == "help" then
		local _addonname = _AddonName:lower()
		PostChat(_format(L["%s commands:"],_AddonName))
		PostChat(_format(L[" /%s - show options"],_addonname))
		PostChat(_format(L[" /%s reset - reset to defaults"],_addonname))
		PostChat(_format(L[" /%s reset [name] - reset to named preset"],_addonname))
		PostChat(_format(L[" /%s help - help message"],_addonname))
	end
end


-----------------------------------------------------------------------------------------------
-- NameView Instance
-----------------------------------------------------------------------------------------------

local NameViewInst = NameView:new()
NameViewInst:Init()

